﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System;
using System.IO;
using System.Net;
using System.Net.Cache;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Threading;
using System.Xml;

namespace LEDControl
{
    public delegate void DataItemRetrieved(string text);

    /// <summary>
    /// Define the types of data we can access, what parameters are needed to access them,
    /// and the function (Get) that retrieves the data
    /// </summary>
    public abstract class DataItemRetriever
    {
        static public Dictionary<string, DataItemRetriever> Templates = new Dictionary<string, DataItemRetriever>();

        static DataItemRetriever()
        {
            Templates["Date/Time"] = new TimeRetriever();
            Templates["XML"] = new XMLRetriever();
            Templates["Text"] = new TextRetriever();
        }

        protected readonly List<Parameter> _parameters = new List<Parameter>();
        public ReadOnlyCollection<Parameter> Parameters
        {
            get { return _parameters.AsReadOnly(); }
        }

        public abstract void BeginGet(IDictionary<string, object> parameters, DataItemRetrieved dataReceived);

        public class Parameter
        {
            public Parameter()
            {
            }

            public Parameter(string name, string description, string toolTip)
                : this(name, description, toolTip, typeof(string))
            {
            }

            public Parameter(string name, string description, string toolTip, Type dataType) : this()
            {
                Name = name;
                Description = description;
                ToolTip = toolTip;
                DataType = dataType;
            }

            public string Name { get; set; }
            public string Description { get; set; }
            public string ToolTip { get; set; }
            public Type DataType { get; set; }
        }

        class TimeRetriever : DataItemRetriever
        {
            public TimeRetriever()
            {
                _parameters.Add(new Parameter("Format", "Date/Time Format", "Use .NET's DateTimeFormatInfo strings."));
            }

            public override void BeginGet(IDictionary<string, object> parameters, DataItemRetrieved dataReceived)
            {
                Application.Current.Dispatcher.BeginInvoke(
                    (Action)
                    delegate
                        {
                            string text =
                                DateTime.Now.ToString(
                                    string.Format(parameters[Parameters[0].Name].ToString()));
                            dataReceived(text);
                        }, DispatcherPriority.Normal);

                return ;
            }
        }

        abstract class URIRetriever : DataItemRetriever
        {
            protected URIRetriever()
            {
                _parameters.Add(new Parameter("URI", "Source URI", "Local file or Internet URI for the XML."));
                _parameters.Add(new Parameter("URI Age", "Max URI Age", "Maximum age of loaded data.", typeof(TimeCountUnits)));
            }

            public override void BeginGet(IDictionary<string, object> parameters, DataItemRetrieved dataReceived)
            {
                new Thread(ReadURI).Start(new object[] {parameters, dataReceived});
            }

            private void ReadURI(object arg)
            {
                try
                {
                    if (arg is object[])
                    {
                        object[] args = (object[]) arg;
                        IDictionary<string, object> parameters = (IDictionary<string, object>) args[0];
                        DataItemRetrieved retrieved = (DataItemRetrieved) args[1];

                        string uri = parameters[Parameters[0].Name].ToString();

                        // See if we have this URI's data cached, and if the cache isn't too
                        // old
                        TimeSpan uriMaxAge = ((TimeCountUnits) parameters[Parameters[1].Name]).ToTimeSpan();

                        // Retrieve data
                        WebRequest req = WebRequest.Create(uri);
                        // Define a cache policy for this request
                        HttpRequestCachePolicy cachePolicy = new HttpRequestCachePolicy(HttpCacheAgeControl.MaxAge,
                                                                                          uriMaxAge);
                        req.CachePolicy = cachePolicy;
                        req.Proxy = null;
                        string text = "";
                        using (WebResponse response = req.GetResponse())
                        {
                            if (response != null)
                            {
                                using (Stream stream = response.GetResponseStream())
                                {
                                    if (stream != null)
                                    {
                                        using (StreamReader reader = new StreamReader(stream))
                                        {
                                            text = reader.ReadToEnd();
                                        }
                                    }
                                }
                            }
                        }

                        text = ExtractText(parameters, text);

                        if (Application.Current != null)
                            Application.Current.Dispatcher.BeginInvoke(
                                (Action) (() => retrieved(text)), DispatcherPriority.Normal);
                    }
                }
                catch
                {
                }
            }

            protected abstract string ExtractText(IDictionary<string, object> parameters, string text);
        }

        class XMLRetriever : URIRetriever
        {
            public XMLRetriever()
            {
                _parameters.Add(new Parameter("XPath", "XPath", "XPath specification to extract the desired value from the XML."));
            }

            protected override string ExtractText(IDictionary<string, object> parameters, string text)
            {
                XmlDocument doc = new XmlDocument();
                doc.LoadXml(text);
                XmlNode node = doc.SelectSingleNode(parameters[Parameters[2].Name].ToString());
                if (node != null)
                    text = node.InnerText;
                return text;
            }
        }

        class TextRetriever : URIRetriever
        {
            public TextRetriever()
            {
                _parameters.Add(new Parameter("Regex", "Regular Expression", ".NET (or Perl) regular expression with a group that identifies the text to return."));
            }

            protected override string ExtractText(IDictionary<string, object> parameters, string text)
            {
                string pattern = parameters[Parameters[2].Name].ToString();
                Match match = Regex.Match(text, pattern);
                if (match.Groups.Count > 1)
                    return match.Groups[1].ToString();

                return "";
            }
        }
    }

    public class ParameterValueUISelector : DataTemplateSelector
    {
        public override DataTemplate SelectTemplate(object item, DependencyObject container)
        {
            if (item is DataItemRetriever.Parameter)
            {
                DataItemRetriever.Parameter parameter = (DataItemRetriever.Parameter)item;
                if (parameter.DataType == typeof(string))
                    return StringTemplate;
                if (parameter.DataType == typeof(TimeCountUnits))
                    return TimeCountUnitsTemplate;
            }
            return base.SelectTemplate(item, container);
        }

        public DataTemplate StringTemplate { get; set; }
        public DataTemplate TimeCountUnitsTemplate { get; set; }
    }
}
