﻿using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Threading;
using Microsoft.Win32.SafeHandles;
using Application = System.Windows.Application;

namespace LEDControl
{
    public delegate void DataReceivedHandler(byte[] data);
    public delegate void DeviceDisconnectedHandler();

    /// <summary>
    /// Low-level HID device control
    /// </summary>
    public class HIDDeviceControl
    {
        private HIDDeviceControl(ushort vendorID, ushort productID)
        {
            dispatcher = Application.Current.Dispatcher;
            deviceID = string.Format("vid_{0:x4}&pid_{1:x4}", vendorID, productID);
            IsOpen = true;
        }

        public bool IsOpen { get; private set; }

        public static HIDDeviceControl Open(ushort vendorID, ushort productID)
        {
            HIDDeviceControl deviceControl = new HIDDeviceControl(vendorID, productID);

            /* 
            Before we can "connect" our application to our USB embedded device, we must first find the device.
            A USB bus can have many devices simultaneously connected, so somehow we have to find our device, and only
            our device.  This is done with the Vendor ID (VID) and Product ID (PID).  Each USB product line should have
            a unique combination of VID and PID.  

            Microsoft has created a number of functions which are useful for finding plug and play devices.  Documentation
            for each function used can be found in the MSDN library.  We will be using the following functions:

            SetupDiGetClassDevs()					//provided by setupapi.dll, which comes with Windows
            SetupDiEnumDeviceInterfaces()			//provided by setupapi.dll, which comes with Windows
            GetLastError()							//provided by kernel32.dll, which comes with Windows
            SetupDiDestroyDeviceInfoList()			//provided by setupapi.dll, which comes with Windows
            SetupDiGetDeviceInterfaceDetail()		//provided by setupapi.dll, which comes with Windows
            SetupDiGetDeviceRegistryProperty()		//provided by setupapi.dll, which comes with Windows
            malloc()								//part of C runtime library, msvcrt.dll?
            CreateFile()							//provided by kernel32.dll, which comes with Windows

            We will also be using the following unusual data types and structures.  Documentation can also be found in
            the MSDN library:

            PSP_DEVICE_INTERFACE_DATA
            PSP_DEVICE_INTERFACE_DETAIL_DATA
            SP_DEVINFO_DATA
            HDEVINFO
            HANDLE
            GUID

            The ultimate objective of the following code is to call CreateFile(), which opens a communications
            pipe to a specific device (such as a HID class USB device endpoint).  CreateFile() returns a "handle" 
            which is needed later when calling ReadFile() or WriteFile().  These functions are used to actually 
            send and receive application related data to/from the USB peripheral device.

            However, in order to call CreateFile(), we first need to get the device path for the USB device
            with the correct VID and PID.  Getting the device path is a multi-step round about process, which
            requires calling several of the SetupDixxx() functions provided by setupapi.dll.
            */

            //Globally Unique Identifier (GUID) for HID class devices.  Windows uses GUIDs to identify things.
            Guid interfaceClassGuid = new Guid(0x4d1e55b2, 0xf16f, 0x11cf, 0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00,
                                               0x30);

            SP_DEVICE_INTERFACE_DATA interfaceDataStructure = new SP_DEVICE_INTERFACE_DATA();
            interfaceDataStructure.cbSize = (uint) Marshal.SizeOf(interfaceDataStructure);

            SP_DEVINFO_DATA devInfoData = new SP_DEVINFO_DATA();

            //First populate a list of plugged in devices (by specifying "DIGCF_PRESENT"), which are of the specified class GUID. 
            IntPtr deviceInfoTable = SetupDiGetClassDevs(ref interfaceClassGuid, null, IntPtr.Zero,
                                                         (uint)
                                                         (DiGetClassFlags.DIGCF_PRESENT |
                                                          DiGetClassFlags.DIGCF_DEVICEINTERFACE));

            //Now look through the list we just populated.  We are trying to see if any of them match our device. 
            uint interfaceIndex = 0;
            while (true)
            {
                if (SetupDiEnumDeviceInterfaces(deviceInfoTable, IntPtr.Zero, ref interfaceClassGuid, interfaceIndex,
                                                ref interfaceDataStructure))
                {
                    int errorStatus = Marshal.GetLastWin32Error();
                    if (ERROR_NO_MORE_ITEMS == errorStatus)
                        //Did we reach the end of the list of matching devices in the deviceInfoTable?
                    {
                        // Could not find the device.  Must not have been attached.
                        SetupDiDestroyDeviceInfoList(deviceInfoTable); // Clean up the old structure we no longer need.
                        return null;
                    }
                }
                else //Else some other kind of unknown error ocurred...
                {
                    SetupDiDestroyDeviceInfoList(deviceInfoTable); // Clean up the old structure we no longer need.
                    return null;
                }

                //Now retrieve the hardware ID from the registry.  The hardware ID contains the VID and PID, which we will then 
                //check to see if it is the correct device or not.

                //Initialize an appropriate SP_DEVINFO_DATA structure.  We need this structure for SetupDiGetDeviceRegistryProperty().
                devInfoData.cbSize = (uint) Marshal.SizeOf(devInfoData);
                SetupDiEnumDeviceInfo(deviceInfoTable, interfaceIndex, ref devInfoData);

                //First query for the size of the hardware ID, so we can know how big a buffer to allocate for the data.
                uint regType, regSize;
                SetupDiGetDeviceRegistryProperty(deviceInfoTable, ref devInfoData, (uint) SPDRP.SPDRP_HARDWAREID,
                                                 out regType, null, 0, out regSize);

                // Allocate a buffer for the hardware ID.
                byte[] propertyValueBuffer = new byte[regSize];

                //Retrieve the hardware IDs for the current device we are looking at.  propertyValueBuffer gets filled with a 
                //REG_MULTI_SZ (array of null terminated strings).  To find a device, we only care about the very first string in the
                //buffer, which will be the "device ID".  The device ID is a string which contains the VID and PID, in the example 
                //format "Vid_04d8&Pid_fbd4".
                SetupDiGetDeviceRegistryProperty(deviceInfoTable, ref devInfoData, (uint) SPDRP.SPDRP_HARDWAREID,
                                                 out regType, propertyValueBuffer, regSize, out regSize);

                //Now check if the first string in the hardware ID matches the device ID of my USB device.
                string deviceIDFromRegistry = Encoding.Unicode.GetString(propertyValueBuffer);

                //Convert to lower case.  This makes the code more robust/portable accross OS Versions
                deviceIDFromRegistry.ToLower();

                //Now check if the hardware ID we are looking at contains the correct VID/PID
                bool matchFound = deviceIDFromRegistry.ToLower().IndexOf(deviceControl.deviceID) != -1;
                if (matchFound)
                {
                    //Device must have been found.  Open read and write handles.  In order to do this, we will need the actual device path first.
                    //We can get the path by calling SetupDiGetDeviceInterfaceDetail(), however, we have to call this function twice:  The first
                    //time to get the size of the required structure/buffer to hold the detailed interface data, then a second time to actually 
                    //get the structure (after we have allocated enough memory for the structure.)
                    //First call populates "structureSize" with the correct value
                    SP_DEVICE_INTERFACE_DETAIL_DATA detailedinterfaceDataStructure =
                        new SP_DEVICE_INTERFACE_DETAIL_DATA();
                    if (IntPtr.Size == 8) // for 64 bit operating systems
                        detailedinterfaceDataStructure.cbSize = 8;
                    else
                        detailedinterfaceDataStructure.cbSize =
                            (uint) (Marshal.SizeOf(detailedinterfaceDataStructure.cbSize) + Marshal.SystemDefaultCharSize);
                    uint structureSize;
                    SetupDiGetDeviceInterfaceDetail(deviceInfoTable, ref interfaceDataStructure,
                                                    IntPtr.Zero, 0, out structureSize,
                                                    IntPtr.Zero);
                    //Now call SetupDiGetDeviceInterfaceDetail() to receive the goods.  
                    SetupDiGetDeviceInterfaceDetail(deviceInfoTable, ref interfaceDataStructure,
                                                    ref detailedinterfaceDataStructure, structureSize, IntPtr.Zero,
                                                    IntPtr.Zero);

                    SetupDiDestroyDeviceInfoList(deviceInfoTable); //Clean up the old structure we no longer need.

                    //We now have the proper device path, and we can finally open read and write handles to the device.
                    //We store the handles in the global variables "writeHandle" and "readHandle", which we will use later to actually communicate.
                    SafeFileHandle writeHandle = CreateFile(detailedinterfaceDataStructure.DevicePath, EFileAccess.GenericWrite,
                                             EFileShare.Read | EFileShare.Write, IntPtr.Zero,
                                             ECreationDisposition.OpenExisting, 0, IntPtr.Zero);

                    deviceControl.writeStream = new FileStream(writeHandle, FileAccess.ReadWrite, 65);
                    
                    SafeFileHandle readHandle = CreateFile(detailedinterfaceDataStructure.DevicePath, 
                        EFileAccess.GenericRead,
                        EFileShare.Read | EFileShare.Write, IntPtr.Zero,
                        ECreationDisposition.OpenExisting,
                        EFileAttributes.Overlapped, IntPtr.Zero);

                    deviceControl.readStream = new FileStream(readHandle, FileAccess.Read, 65, true);

                    // Start an asynchronous read operation
                    deviceControl.StartAsynchronousRead();

                    return deviceControl;
                }

                interfaceIndex++;
                //Keep looping until we either find a device with matching VID and PID, or until we run out of items.
            } //end of while(true)	
        }

        public void Write(byte endpoint, byte[] output)
        {
            if (writeStream != null)
            {
                if (output.Length > 64)
                    throw new ArgumentException("output must be less than 65 bytes long.");
                byte[] outputBuffer = new byte[65];
                outputBuffer[0] = endpoint;
                Array.Copy(output, 0, outputBuffer, 1, output.Length);

                try
                {
                    writeStream.Write(outputBuffer, 0, 65);
                }
                catch
                {
                    Close();
                }
            }
        }

        private void StartAsynchronousRead()
        {
            readStream.BeginRead(inputBuffer, 0, inputBuffer.Length, OnReadCompleted, null);
        }

        private void OnReadCompleted(IAsyncResult ar)
        {
            try
            {
                if (readStream != null)
                {
                    readStream.EndRead(ar);

                    dispatcher.BeginInvoke(
                        (Action) delegate
                                     {
                                         if (DataReceived != null)
                                             DataReceived(inputBuffer);

                                         StartAsynchronousRead();
                                     }, DispatcherPriority.Normal);
                }
            }
            catch (IOException)
            {
                dispatcher.BeginInvoke(
                    (Action) Close, DispatcherPriority.Normal);
            }
        }

        private void Close()
        {
            readStream.Close();
            readStream = null;
            writeStream.Close();
            writeStream = null;
            IsOpen = false;
        }

        #region Fields

        private readonly byte[] inputBuffer = new byte[65];
        private FileStream readStream;
        private FileStream writeStream;
        private readonly string deviceID;
        private readonly Dispatcher dispatcher;

        #endregion

        #region Events

        public event DataReceivedHandler DataReceived;

        #endregion

        #region Windows APIs

        [DllImport("setupapi.dll", CharSet = CharSet.Auto)]
        static extern IntPtr SetupDiGetClassDevs(
            ref Guid ClassGuid,
            [MarshalAs(UnmanagedType.LPTStr)] string Enumerator,
            IntPtr hwndParent,
            UInt32 Flags
            );

        [FlagsAttribute]
        public enum DiGetClassFlags : uint
        {
            DIGCF_DEFAULT       = 0x00000001,  // only valid with DIGCF_DEVICEINTERFACE
            DIGCF_PRESENT       = 0x00000002,
            DIGCF_ALLCLASSES    = 0x00000004,
            DIGCF_PROFILE       = 0x00000008,
            DIGCF_DEVICEINTERFACE   = 0x00000010,
        }

        [DllImport(@"setupapi.dll", CharSet=CharSet.Auto, SetLastError = true)]
        public static extern Boolean SetupDiEnumDeviceInterfaces(
            IntPtr hDevInfo,
            IntPtr devInfo,
            ref Guid interfaceClassGuid,
            UInt32 memberIndex,
            ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData
            );

        [StructLayout(LayoutKind.Sequential, Pack = 2)]
        public struct SP_DEVICE_INTERFACE_DATA
        {
            public uint cbSize;
            public Guid InterfaceClassGuid;
            public uint Flags;
            public IntPtr Reserved;    
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct SP_DEVINFO_DATA
        {
            public uint cbSize;
            public Guid ClassGuid;
            public uint DevInst;
            public IntPtr Reserved;
        }

        /// <summary>
        /// No more data is available.
        /// </summary>
        public const int ERROR_NO_MORE_ITEMS = 259;

        [DllImport(@"setupapi.dll", CharSet=CharSet.Auto, SetLastError = true)]
        public static extern bool SetupDiDestroyDeviceInfoList( IntPtr hDevInfo );

        [DllImport("setupapi.dll", SetLastError=true)]
        static extern bool SetupDiEnumDeviceInfo(IntPtr DeviceInfoSet, uint MemberIndex, ref SP_DEVINFO_DATA DeviceInfoData);
        
        [DllImport("setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern bool SetupDiGetDeviceRegistryProperty(
            IntPtr DeviceInfoSet,
            ref SP_DEVINFO_DATA DeviceInfoData,
            uint Property,
            out UInt32 PropertyRegDataType,
            byte[] PropertyBuffer,
            uint PropertyBufferSize,
            out UInt32 RequiredSize
            );
        
        /// <summary>
        /// Device registry property codes
        /// </summary>
        public enum SPDRP
        {
            /// <summary>
            /// DeviceDesc (R/W)
            /// </summary>
            SPDRP_DEVICEDESC = 0x00000000,

            /// <summary>
            /// HardwareID (R/W)
            /// </summary>
            SPDRP_HARDWAREID = 0x00000001,

            /// <summary>
            /// CompatibleIDs (R/W)
            /// </summary>
            SPDRP_COMPATIBLEIDS = 0x00000002,

            /// <summary>
            /// unused
            /// </summary>
            SPDRP_UNUSED0 = 0x00000003,

            /// <summary>
            /// Service (R/W)
            /// </summary>
            SPDRP_SERVICE = 0x00000004,

            /// <summary>
            /// unused
            /// </summary>
            SPDRP_UNUSED1 = 0x00000005,

            /// <summary>
            /// unused
            /// </summary>
            SPDRP_UNUSED2 = 0x00000006,

            /// <summary>
            /// Class (R--tied to ClassGUID)
            /// </summary>
            SPDRP_CLASS = 0x00000007,

            /// <summary>
            /// ClassGUID (R/W)
            /// </summary>
            SPDRP_CLASSGUID = 0x00000008,

            /// <summary>
            /// Driver (R/W)
            /// </summary>
            SPDRP_DRIVER = 0x00000009,

            /// <summary>
            /// ConfigFlags (R/W)
            /// </summary>
            SPDRP_CONFIGFLAGS = 0x0000000A,

            /// <summary>
            /// Mfg (R/W)
            /// </summary>
            SPDRP_MFG = 0x0000000B,

            /// <summary>
            /// FriendlyName (R/W)
            /// </summary>
            SPDRP_FRIENDLYNAME = 0x0000000C,

            /// <summary>
            /// LocationInformation (R/W)
            /// </summary>
            SPDRP_LOCATION_INFORMATION = 0x0000000D,

            /// <summary>
            /// PhysicalDeviceObjectName (R)
            /// </summary>
            SPDRP_PHYSICAL_DEVICE_OBJECT_NAME = 0x0000000E,

            /// <summary>
            /// Capabilities (R)
            /// </summary>
            SPDRP_CAPABILITIES = 0x0000000F,

            /// <summary>
            /// UiNumber (R)
            /// </summary>
            SPDRP_UI_NUMBER = 0x00000010,

            /// <summary>
            /// UpperFilters (R/W)
            /// </summary>
            SPDRP_UPPERFILTERS = 0x00000011,

            /// <summary>
            /// LowerFilters (R/W)
            /// </summary>
            SPDRP_LOWERFILTERS = 0x00000012,

            /// <summary>
            /// BusTypeGUID (R)
            /// </summary>
            SPDRP_BUSTYPEGUID = 0x00000013,

            /// <summary>
            /// LegacyBusType (R)
            /// </summary>
            SPDRP_LEGACYBUSTYPE = 0x00000014,

            /// <summary>
            /// BusNumber (R)
            /// </summary>
            SPDRP_BUSNUMBER = 0x00000015,

            /// <summary>
            /// Enumerator Name (R)
            /// </summary>
            SPDRP_ENUMERATOR_NAME = 0x00000016,

            /// <summary>
            /// Security (R/W, binary form)
            /// </summary>
            SPDRP_SECURITY = 0x00000017,

            /// <summary>
            /// Security (W, SDS form)
            /// </summary>
            SPDRP_SECURITY_SDS = 0x00000018,

            /// <summary>
            /// Device Type (R/W)
            /// </summary>
            SPDRP_DEVTYPE = 0x00000019,

            /// <summary>
            /// Device is exclusive-access (R/W)
            /// </summary>
            SPDRP_EXCLUSIVE = 0x0000001A,

            /// <summary>
            /// Device Characteristics (R/W)
            /// </summary>
            SPDRP_CHARACTERISTICS = 0x0000001B,

            /// <summary>
            /// Device Address (R)
            /// </summary>
            SPDRP_ADDRESS = 0x0000001C,

            /// <summary>
            /// UiNumberDescFormat (R/W)
            /// </summary>
            SPDRP_UI_NUMBER_DESC_FORMAT = 0X0000001D,

            /// <summary>
            /// Device Power Data (R)
            /// </summary>
            SPDRP_DEVICE_POWER_DATA = 0x0000001E,

            /// <summary>
            /// Removal Policy (R)
            /// </summary>
            SPDRP_REMOVAL_POLICY = 0x0000001F,

            /// <summary>
            /// Hardware Removal Policy (R)
            /// </summary>
            SPDRP_REMOVAL_POLICY_HW_DEFAULT = 0x00000020,

            /// <summary>
            /// Removal Policy Override (RW)
            /// </summary>
            SPDRP_REMOVAL_POLICY_OVERRIDE = 0x00000021,

            /// <summary>
            /// Device Install State (R)
            /// </summary>
            SPDRP_INSTALL_STATE = 0x00000022,

            /// <summary>
            /// Device Location Paths (R)
            /// </summary>
            SPDRP_LOCATION_PATHS = 0x00000023,
        }

        [DllImport(@"setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern Boolean SetupDiGetDeviceInterfaceDetail(
           IntPtr hDevInfo,
           ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData,
           ref SP_DEVICE_INTERFACE_DETAIL_DATA deviceInterfaceDetailData,
           UInt32 deviceInterfaceDetailDataSize,
           IntPtr requiredSize,
           IntPtr deviceInfoData
        );

        [DllImport(@"setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern Boolean SetupDiGetDeviceInterfaceDetail(
           IntPtr hDevInfo,
           ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData,
           IntPtr deviceInterfaceDetailData,
           UInt32 deviceInterfaceDetailDataSize,
           out UInt32 requiredSize,
           IntPtr deviceInfoData
        );

        // Device interface detail data
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        public struct SP_DEVICE_INTERFACE_DETAIL_DATA
        {
            public UInt32 cbSize;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 256)]
            public string DevicePath;
        }

        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        public static extern SafeFileHandle CreateFile(
           string lpFileName,
           EFileAccess dwDesiredAccess,
           EFileShare dwShareMode,
           IntPtr lpSecurityAttributes,
           ECreationDisposition dwCreationDisposition,
           EFileAttributes dwFlagsAndAttributes,
           IntPtr hTemplateFile);
        [Flags]
        public enum EFileAccess : uint
        {
            /// <summary>
            ///
            /// </summary>
            GenericRead = 0x80000000,
            /// <summary>
            ///
            /// </summary>
            GenericWrite = 0x40000000,
            /// <summary>
            ///
            /// </summary>
            GenericExecute = 0x20000000,
            /// <summary>
            ///
            /// </summary>
            GenericAll = 0x10000000
        }

        [Flags]
        public enum EFileShare : uint
        {
            /// <summary>
            ///
            /// </summary>
            None = 0x00000000,
            /// <summary>
            /// Enables subsequent open operations on an object to request read access.
            /// Otherwise, other processes cannot open the object if they request read access.
            /// If this flag is not specified, but the object has been opened for read access, the function fails.
            /// </summary>
            Read = 0x00000001,
            /// <summary>
            /// Enables subsequent open operations on an object to request write access.
            /// Otherwise, other processes cannot open the object if they request write access.
            /// If this flag is not specified, but the object has been opened for write access, the function fails.
            /// </summary>
            Write = 0x00000002,
            /// <summary>
            /// Enables subsequent open operations on an object to request delete access.
            /// Otherwise, other processes cannot open the object if they request delete access.
            /// If this flag is not specified, but the object has been opened for delete access, the function fails.
            /// </summary>
            Delete = 0x00000004
        }

        public enum ECreationDisposition : uint
        {
            /// <summary>
            /// Creates a new file. The function fails if a specified file exists.
            /// </summary>
            New = 1,
            /// <summary>
            /// Creates a new file, always.
            /// If a file exists, the function overwrites the file, clears the existing attributes, combines the specified file attributes,
            /// and flags with FILE_ATTRIBUTE_ARCHIVE, but does not set the security descriptor that the SECURITY_ATTRIBUTES structure specifies.
            /// </summary>
            CreateAlways = 2,
            /// <summary>
            /// Opens a file. The function fails if the file does not exist.
            /// </summary>
            OpenExisting = 3,
            /// <summary>
            /// Opens a file, always.
            /// If a file does not exist, the function creates a file as if dwCreationDisposition is CREATE_NEW.
            /// </summary>
            OpenAlways = 4,
            /// <summary>
            /// Opens a file and truncates it so that its size is 0 (zero) bytes. The function fails if the file does not exist.
            /// The calling process must open the file with the GENERIC_WRITE access right.
            /// </summary>
            TruncateExisting = 5
        }

        [Flags]
        public enum EFileAttributes : uint
        {
            Readonly = 0x00000001,
            Hidden = 0x00000002,
            System = 0x00000004,
            Directory = 0x00000010,
            Archive = 0x00000020,
            Device = 0x00000040,
            Normal = 0x00000080,
            Temporary = 0x00000100,
            SparseFile = 0x00000200,
            ReparsePoint = 0x00000400,
            Compressed = 0x00000800,
            Offline = 0x00001000,
            NotContentIndexed = 0x00002000,
            Encrypted = 0x00004000,
            Write_Through = 0x80000000,
            Overlapped = 0x40000000,
            NoBuffering = 0x20000000,
            RandomAccess = 0x10000000,
            SequentialScan = 0x08000000,
            DeleteOnClose = 0x04000000,
            BackupSemantics = 0x02000000,
            PosixSemantics = 0x01000000,
            OpenReparsePoint = 0x00200000,
            OpenNoRecall = 0x00100000,
            FirstPipeInstance = 0x00080000
        }

        #endregion

    }
}
