﻿using System;
using System.ComponentModel;
using System.Xml.Serialization;

namespace LEDControl
{
    /// <summary>
    /// The life cycle of an application is:
    ///    Create()     (LEDApplication object is created)
    ///    Start()      (User has chosen to run the app)
    ///    Activate()   (The app is visible)
    ///    Deactivate() (The app is covered up or about to terminate)
    ///    Terminate()  (The app is no longer running)
    /// 
    /// An LEDApplication is created once, can be Start/Terminated multiple times.
    /// Between Start and Terminate, it can be Acivate/Deactivated multiple times.
    /// </summary>
    [XmlInclude(typeof(LEDApplicationSelector))]
    [XmlInclude(typeof(LEDApplicationLauncher))]
    [XmlInclude(typeof(LEDApplicationMenu))]
    [XmlInclude(typeof(LEDApplicationDataViewer))]
    public class LEDApplication : INotifyPropertyChanged
    {
        #region Fields

        public static readonly string[] ApplicationTypes = {"Menu", "Launcher", "Data Viewer"};

        #endregion

        #region Construction

        public static LEDApplication Create(string typeName, string name)
        {
            LEDApplication ledApplication;

            if (typeName == ApplicationTypes[0])
                ledApplication = new LEDApplicationMenu();
            else if (typeName == ApplicationTypes[1])
                ledApplication = new LEDApplicationLauncher();
            else if (typeName == ApplicationTypes[2])
                ledApplication = new LEDApplicationDataViewer(true);
            else
                throw new ArgumentException("Unable to create object of type '" + typeName + "'");

            ledApplication.ApplicationType = typeName;
            ledApplication.Name = name;

            return ledApplication;
        }

        #endregion

        #region Lifecycle control

        public virtual bool Start(LEDApplicationTree tree, LifecycleArgs args)
        {
            IsStarted = true;
            return true;
        }

        /// <summary>
        /// Called when this application becomes the currently running application (either
        /// because it's been added to the running stack or a item on top of it has been
        /// shutdown)
        /// </summary>
        public virtual void Activate(LEDApplicationTree tree, LifecycleArgs args)
        {
            IsActivated = true;
            _ledDevice = new WeakReference(args.LEDDevice);
        }

        /// <summary>
        /// Called when this application is no longer active (either because it's
        /// been covered by another application, in which case it will be activated again
        /// when that app terminates) or it's being terminated
        /// </summary>
        public virtual void Deactivate(LEDApplicationTree tree, LifecycleArgs args)
        {
            IsActivated = false;
            _ledDevice = null;
        }

        public virtual void Terminate(LEDApplicationTree tree, LifecycleArgs args)
        {
            IsStarted = false;
        }

        #endregion

        public void Refresh(LEDDevice device)
        {
            device.Refresh();
        }

        #region Services for Subclasses

        protected void RequestExecute(Runner runner, LEDApplicationTree ledApplicationTree)
        {
            runner.RequestExecute(ledApplicationTree);
        }

        protected internal void RequestTerminate(Runner runner)
        {
            runner.RequestTerminate(this);
        }

        #endregion

        protected virtual int GetMaxChildren() 
        {
            return 0;
        }

        #region Device Event Handlers

        public virtual void OnPreviewButtonDown(LEDApplicationTree node, LEDDeviceButtonEventArgs args)
        {
        }

        public virtual void OnPreviewButtonUp(LEDApplicationTree node, LEDDeviceButtonEventArgs args)
        {
        }

        public virtual void OnButtonDown(LEDApplicationTree node, LEDDeviceButtonEventArgs args)
        {
        }

        public virtual void OnButtonUp(LEDApplicationTree node, LEDDeviceButtonEventArgs args)
        {
        }

        public virtual void OnPreviewAmbientLightChanged(LEDApplicationTree node, LEDDeviceAmbientLightEventArgs args)
        {
        }

        public virtual void OnAmbientLightChanged(LEDApplicationTree node, LEDDeviceAmbientLightEventArgs args)
        {
        }

        #endregion

        #region Properties

        private string _name;

        public string Name
        {
            get { return _name; }
            set
            {
                _name = value;
                OnPropertyChanged("Name");
            }
        }

        [XmlIgnore]
        public int MaxChildren { get { return GetMaxChildren(); } }

        public string ApplicationType { get; set; }

        protected LEDDevice LEDDevice
        {
            get
            {
                if (_ledDevice != null)
                    return _ledDevice.Target as LEDDevice;
                return null;
            }
        }

        protected bool IsActivated { get; private set; }
        protected bool IsStarted { get; private set; }

        #endregion

        #region Fields

        private WeakReference _ledDevice;

        #endregion

        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string name)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(name));
        }

        #endregion
    }
}
