﻿
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.ComponentModel;
using System.Linq.Dynamic;
using System.Media;
using System.Windows;
using System.Xml.Serialization;
using System.Windows.Threading;
using System.Linq.Expressions;

namespace LEDControl
{
    public class LEDApplicationDataViewer : LEDApplication
    {
        /// <summary>
        /// Called when we're creating this object from XML stream
        /// </summary>
        public LEDApplicationDataViewer()
        {
            DataItems = new ObservableCollection<DataItem>();
            Format = "{0}";
            IndicatorExpressions = new ObservableCollection<Expression>();
            ButtonFunctions = new ObservableCollection<ButtonFunction>();
            ActionExpression = new Expression();
        }

        /// <summary>
        /// Called when we're creating this object from code
        /// </summary>
        /// <param name="init"></param>
        public LEDApplicationDataViewer(bool init)
            : this()
        {
            // If we should fill the fixed-length arrays, do so
            if (init)
            {
                for (int index = 0; index < 4; index++)
                    IndicatorExpressions.Add(new Expression { Text = "false" });

                for (int index = 0; index < 8; index++)
                    ButtonFunctions.Add(new ButtonFunction());
                ButtonFunctions[0].Action = ButtonFunction.ButtonAction.Exit;
            }
        }

        public void AddDataItem()
        {
            DataItem dataItem = new DataItem
                                {
                                    TemplateName = DataItemRetriever.Templates.Keys.ElementAt(0)
                                };
            DataItems.Add(dataItem);
        }

        public void DeleteDataItem(DataItem toBeDeleted)
        {
            DataItems.Remove(toBeDeleted);
        }

        public void Test()
        {
            foreach (DataItem dataItem in DataItems)
                dataItem.RefreshValue();
        }

        protected override int GetMaxChildren()
        {
            return 0;
        }

        public override void Activate(LEDApplicationTree tree, LifecycleArgs args)
        {
            base.Activate(tree, args);

            foreach (DataItem dataItem in DataItems)
                dataItem.StartTimer();
            WriteValue();
        }

        public override void Deactivate(LEDApplicationTree tree, LifecycleArgs args)
        {
            LEDDevice.SetIndicators(new Indicators());
            base.Deactivate(tree, args);

            foreach (DataItem dataItem in DataItems)
                dataItem.StopTimer();
        }

        /// <summary>
        /// Output the calculated value of the DataItems combined into the Format.
        /// Also set the indicators
        /// </summary>
        private void WriteValue()
        {
            string displayValue = DisplayValue;

            // Evaluate all the indicator expressions
            for (int index = 0; index < TestIndicatorValues.Length; index++)
            {
                TestIndicatorValues[index] = false;
                if (index < IndicatorExpressions.Count)
                {
                    string indicatorExpression = IndicatorExpressions[index].Text;
                    if (string.IsNullOrEmpty(IndicatorExpressions[index].Error) && !string.IsNullOrEmpty(indicatorExpression))
                    {
                        try
                        {
                            // Convert {n} into the value of the n-th data item
                            indicatorExpression = string.Format(indicatorExpression,
                                                                DataItems.Select(item => (object) SafeString(item.Value)).ToArray());
                            // Convert ^n^ into the previous value of the n-th data item
                            for (int i = 0;  i < DataItems.Count; ++i)
                                indicatorExpression = indicatorExpression.Replace("^" + i + "^", SafeString(DataItems[i].PreviousValue));

                            LambdaExpression lambda = DynamicExpression.ParseLambda(
                                new ParameterExpression[0], typeof(bool), indicatorExpression);
                            TestIndicatorValues[index] = (bool)lambda.Compile().DynamicInvoke(null);
                        }
                        catch
                        {
                            // If none of the data values were null (which they might be before we've evaluated
                            // each one at least once)
                            if (DataItems.FirstOrDefault(item => item.Value == null) == null)
                                // Then the error must be due to bad syntax
                                IndicatorExpressions[index].Error = "Unable to execute \"" + indicatorExpression + "\"";
                        }
                    }
                }
            }

            // Evaluate the action expressions
            TestActionValue = null;
            if (string.IsNullOrEmpty(ActionExpression.Error) && !string.IsNullOrEmpty(ActionExpression.Text))
            {
                string expression = ActionExpression.Text;
                try
                {
                    // Convert {n} into the value of the n-th data item
                    expression = string.Format(expression,
                                               DataItems.Select(item => (object)SafeString(item.Value)).ToArray());
                    // Convert ^n^ into the previous value of the n-th data item
                    for (int i = 0; i < DataItems.Count; ++i)
                        expression = expression.Replace("^" + i + "^", SafeString(DataItems[i].PreviousValue));

                    if (expression != _lastActionExpression)
                    {
                        _lastActionExpression = expression;
                        LambdaExpression lambda = DynamicExpression.ParseLambda(
                            new ParameterExpression[0], typeof(string), expression);
                        TestActionValue = (string) lambda.Compile().DynamicInvoke(null);
                    }
                }
                catch
                {
                    // If none of the data values were null (which they might be before we've evaluated
                    // each one at least once)
                    if (DataItems.FirstOrDefault(item => item.Value == null) == null)
                        // Then the error must be due to bad syntax
                        ActionExpression.Error = "Unable to execute \"" + expression + "\"";
                }
            }

            if (IsActivated)
            {
                LEDDevice ledDevice = LEDDevice;
                ledDevice.Write(displayValue);
                ledDevice.SetIndicators(
                    new Indicators
                    {
                        TopLeft = TestIndicatorValues[0],
                        BottomLeft = TestIndicatorValues[1],
                        TopRight = TestIndicatorValues[2],
                        BottomRight = TestIndicatorValues[3]
                    });

                if (!string.IsNullOrEmpty(TestActionValue))
                {
                    if (File.Exists(TestActionValue))
                    {
                        try
                        {
                            using (SoundPlayer player = new SoundPlayer(TestActionValue))
                                player.Play();
                        }
                        catch {}
                    }
                }
            }
            TestValue = displayValue;
            OnPropertyChanged("TestIndicatorValues");
        }

        private static string SafeString(string value)
        {
            return value == null ? value : value.Replace(@"\", "").Replace("\"", "");
        }

        public override void OnButtonDown(LEDApplicationTree node, LEDDeviceButtonEventArgs args)
        {
            for (int i = 0; i < ButtonFunctions.Count; ++i)
            {
                if (((int) args.Button & (1 << i)) != 0)
                {
                    switch (ButtonFunctions[i].Action)
                    {
                        case ButtonFunction.ButtonAction.Exit :
                            RequestTerminate(args.Runner);
                            args.Handled = true;
                            return;

                        case ButtonFunction.ButtonAction.Execute :
                            if (!string.IsNullOrEmpty(ButtonFunctions[i].Argument))
                            {
                                Process process = new Process
                                                  {
                                                      StartInfo =
                                                          {
                                                              FileName = ButtonFunctions[i].Argument,
                                                              UseShellExecute = true
                                                          }
                                                  };
                                process.Start();
                            }
                            break;

                        default :
                            break;
                    }
                }
            }
        }

        #region Properties

        private ObservableCollection<DataItem> _dataItems;
        public ObservableCollection<DataItem> DataItems
        {
            get { return _dataItems; }
            set
            {
                if (_dataItems != null)
                    _dataItems.CollectionChanged -= _dataItems_CollectionChanged;
                _dataItems = value;
                if (_dataItems != null)
                    _dataItems.CollectionChanged += _dataItems_CollectionChanged;
                OnPropertyChanged("DataItems");
            }
        }

        private string _format;
        public string Format
        {
            get { return _format; }
            set { _format = value; OnPropertyChanged("Format"); }
        }

        private ObservableCollection<Expression> _indicatorExpressions;
        public ObservableCollection<Expression> IndicatorExpressions
        {
            get { return _indicatorExpressions; }
            set
            {
                _indicatorExpressions = value;
                _indicatorExpressions.CollectionChanged += _indicatorExpressions_CollectionChanged;
            }
        }

        void _indicatorExpressions_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (object t in e.NewItems)
                {
                    ((INotifyPropertyChanged)t).PropertyChanged += delegate { OnPropertyChanged("IndicatorExpressions"); };
                }
            }
        }

        private Expression _actionExpression;
        public Expression ActionExpression
        {
            get { return _actionExpression; }
            set { _actionExpression = value; OnPropertyChanged("ActionExpression");}
        }

        [XmlIgnore]
        public string DisplayValue
        {
            get
            {
                try
                {
                    return string.Format(Format, DataItems.Select(item => (object)item.Value).ToArray());
                }
                catch
                {
                    return "";
                }
            }
        }

        private string _testValue;

        [XmlIgnore]
        public string TestValue
        {
            get { return _testValue; }
            set { _testValue = value; OnPropertyChanged("TestValue"); }
        }

        bool[] _testIndicatorValues = new bool[4];
        [XmlIgnore]
        public bool[] TestIndicatorValues
        {
            get { return _testIndicatorValues; }
            set { _testIndicatorValues = value; OnPropertyChanged("TestIndicatorValues"); }
        }

        private string _testActionValue;
        [XmlIgnore]
        public string TestActionValue
        {
            get { return _testActionValue; }
            set { _testActionValue = value; OnPropertyChanged("TestActionValue"); }
        }

        private ObservableCollection<ButtonFunction> _buttonFunctions;
        public ObservableCollection<ButtonFunction> ButtonFunctions
        {
            get { return _buttonFunctions; }
            set
            {
                _buttonFunctions = value;
                _buttonFunctions.CollectionChanged += _buttonFunctions_CollectionChanged;
            }
        }

        void _buttonFunctions_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (ButtonFunction buttonFunction in e.NewItems.OfType<ButtonFunction>())
                {
                    buttonFunction.PropertyChanged += delegate { OnPropertyChanged("ButtonFunctions"); };
                }
            }
        }

        private static readonly string[] _indicatorLabels = { "Top Left", "Bottom Left", "Top Right", "Bottom Right" };
        private string _lastActionExpression;
        static public string[] IndicatorLabels { get { return _indicatorLabels; } }

        #endregion

        #region Event Handlers

        void _dataItems_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.OldItems != null)
            {
                foreach (DataItem oldItem in e.OldItems.OfType<DataItem>())
                    oldItem.PropertyChanged -= dataItem_PropertyChanged;
            }
            if (e.NewItems != null)
            {
                foreach (DataItem newItem in e.NewItems.OfType<DataItem>())
                    newItem.PropertyChanged += dataItem_PropertyChanged;
            }
            OnPropertyChanged("DataItems");
        }

        /// <summary>
        /// Called when a DataItem property changes
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void dataItem_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            // If the changed property is Value
            if (e.PropertyName == "Value")
                // Update the LED display, but don't propagate the change to the model,
                // since Value is not persisted
                WriteValue();
            else
                OnPropertyChanged("DataItems");
        }

        #endregion

        public class Expression : INotifyPropertyChanged
        {
            private string _text;
            public string Text
            {
                get { return _text; }
                set
                {
                    _text = value;
                    OnPropertyChanged("Text");
                    Error = null;
                }
            }

            /// <summary>
            /// Set if executing the expression results in an error
            /// </summary>
            private string _error;
            [XmlIgnore]
            public string Error
            {
                get { return _error; }
                set { _error = value; OnPropertyChanged("Error"); }
            }

            #region Implementation of INotifyPropertyChanged

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged(string name)
            {
                if (PropertyChanged != null)
                    PropertyChanged(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }

        public class DataItem : INotifyPropertyChanged
        {
            public DataItem()
            {
                Description = "";
                ParameterValues = new ObservableCollection<ParameterValue>();
                RefreshPeriod = new TimeCountUnits { Count = 1, Units = "Minutes" };
            }

            public void StartTimer()
            {
                StopTimer();
                _timer = new DispatcherTimer(RefreshPeriod.ToTimeSpan(), DispatcherPriority.Render,
                                             RefreshValue,
                                             Application.Current.Dispatcher);

                if (RefreshPeriod.ToTimeSpan().Ticks != 0)
                {
                    _timer.Start();
                    RefreshValue(null, null);
                }
            }

            private void RefreshValue(object sender, EventArgs e)
            {
                RefreshValue();
            }

            internal void RefreshValue()
            {
                Dictionary<string, object> parameters = new Dictionary<string, object>();
                foreach (DataItemRetriever.Parameter parameter in Parameters)
                {
                    ParameterValue value = ParameterValues.FirstOrDefault(item => item.ParameterName == parameter.Name);
                    parameters[parameter.Name] = (value == null ? "" : value.Value);
                }
                DataItemRetriever.Templates[TemplateName].BeginGet(parameters, OnDataRetrieved);
            }

            private void OnDataRetrieved(string text)
            {
                Value = text;
            }

            public void StopTimer()
            {
                if (_timer != null)
                {
                    _timer.Stop();
                    _timer = null;
                }
            }

            #region Properties

            private string _templateName;
            [XmlAttribute]
            public string TemplateName
            {
                get { return _templateName; }
                set
                {
                    _templateName = value;
                    Parameters = DataItemRetriever.Templates[TemplateName].Parameters;
                    OnPropertyChanged("TemplateName");
                    OnPropertyChanged("Parameters");
                }
            }

            private string _description;
            public string Description
            {
                get { return _description; }
                set { _description = value; OnPropertyChanged("Description"); }
            }

            [XmlIgnore]
            public ReadOnlyCollection<DataItemRetriever.Parameter> Parameters { get; set; }

            private ObservableCollection<ParameterValue> _parameterValues;
            public ObservableCollection<ParameterValue> ParameterValues
            {
                get { return _parameterValues; }
                set
                {
                    _parameterValues = value;
                    ParameterValues.CollectionChanged += ParameterValues_CollectionChanged;
                }
            }

            private TimeCountUnits _refreshPeriod;
            public TimeCountUnits RefreshPeriod
            {
                get { return _refreshPeriod; }
                set
                {
                    if (_refreshPeriod != null)
                        _refreshPeriod.PropertyChanged -= _refreshPeriod_PropertyChanged;
                    _refreshPeriod = value;
                    if (_refreshPeriod != null)
                        _refreshPeriod.PropertyChanged += _refreshPeriod_PropertyChanged;
                    if (_timer != null)
                        StartTimer();
                }
            }

            void _refreshPeriod_PropertyChanged(object sender, PropertyChangedEventArgs e)
            {
                if (_timer != null)
                    StartTimer();
            }

            private bool _toUpper;

            [XmlAttribute]
            public bool ToUpper
            {
                get { return _toUpper; }
                set
                {
                    _toUpper = value;
                    OnPropertyChanged("ToUpper");
                }
            }

            /// <summary>
            /// The computed value of this DataItem
            /// </summary>
            string _value;
            [XmlIgnore]
            public string Value
            {
                get { return _value; }
                set
                {
                    PreviousValue = _value;
                    _value = ToUpper ? value.ToUpper() : value;
                    OnPropertyChanged("Value");
                }
            }

            [XmlIgnore]
            public string PreviousValue { get; set; }

            #endregion

            #region Event Handlers

            void ParameterValues_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
            {
                if (e.OldItems != null)
                {
                    foreach (ParameterValue oldItem in e.OldItems.OfType<ParameterValue>())
                        oldItem.PropertyChanged -= PropertyValue_PropertyChanged;
                }
                if (e.NewItems != null)
                {
                    foreach (ParameterValue newItem in e.NewItems.OfType<ParameterValue>())
                        newItem.PropertyChanged += PropertyValue_PropertyChanged;
                }
                OnPropertyChanged("ParameterValues");
            }

            private void PropertyValue_PropertyChanged(object sender, PropertyChangedEventArgs e)
            {
                OnPropertyChanged("ParameterValues");
            }

            #endregion

            #region Fields

            private DispatcherTimer _timer;
            #endregion

            public class ParameterValue : INotifyPropertyChanged
            {
                public string ParameterName { get; set; }

                private object _value;
                [XmlElement("StringValue", typeof(string))]
                [XmlElement("TimeCountUnitsValue", typeof(TimeCountUnits))]
                public object Value
                {
                    get { return _value; }
                    set
                    {
                        if (_value != value)
                        {
                            _value = value;
                            OnPropertyChanged("Value");
                        }
                    }
                }

                public event PropertyChangedEventHandler PropertyChanged;
                void OnPropertyChanged(string name)
                {
                    if (PropertyChanged != null)
                        PropertyChanged(this, new PropertyChangedEventArgs(name));
                }
            }

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged(string name)
            {
                if (PropertyChanged != null)
                    PropertyChanged(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }

        public class ButtonFunction : INotifyPropertyChanged
        {
            public enum ButtonAction
            {
                None,
                Exit,
                Execute
            }

            public static ButtonAction[] ActionChoices = { ButtonAction.None, ButtonAction.Exit, ButtonAction.Execute };

            private ButtonAction _action;
            public ButtonAction Action
            {
                get { return _action; }
                set { _action = value; OnPropertyChanged("Action"); OnPropertyChanged("ArgumentVisibility"); }
            }

            private string _argument;
            public string Argument
            {
                get { return _argument; }
                set { _argument = value; OnPropertyChanged("Argument"); }
            }

            [XmlIgnore]
            public Visibility ArgumentVisibility 
            {
                get
                {
                    return (Action == ButtonAction.Execute
                                ? Visibility.Visible
                                : Visibility.Collapsed);
                }
            }

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged(string name)
            {
                if (PropertyChanged != null)
                    PropertyChanged(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }
    }
}
