﻿using System.Collections.ObjectModel;
using System.Data;
using System.ComponentModel;
using System.Linq;
using System.Xml.Serialization;

namespace LEDControl
{
    public class LEDApplicationTree : INotifyPropertyChanged
    {
        public LEDApplicationTree()
        {
            Children = new ObservableCollection<LEDApplicationTree>();
        }

        public LEDApplicationTree(LEDApplication application) : this()
        {
            Application = application;
        }

        public void AddChild(LEDApplicationTree child)
        {
            if (_children.Count < Application.MaxChildren)
                _children.Add(child);
            else
                throw new ConstraintException("Application " + Application.Name + " can not accept more children.");
        }

        #region Event Handlers

        void _application_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName != "TestValue" && e.PropertyName != "TestIndicatorValues")
                OnPropertyChanged("Application");
        }

        /// <summary>
        /// When our list of children changes, make sure their Parent property is set correctly
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _children_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.OldItems != null)
            {
                foreach (LEDApplicationTree oldItem in e.OldItems.OfType<LEDApplicationTree>())
                {
                    oldItem.Parent = null;
                    oldItem.PropertyChanged -= child_PropertyChanged;
                }
            }
            foreach (LEDApplicationTree child in Children)
            {
                child.Parent = this;
                child.PropertyChanged += child_PropertyChanged;
            }
            
            OnPropertyChanged("Children");
        }

        void child_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            OnPropertyChanged("Children");
        }

        #endregion

        #region Properties

        private LEDApplication _application;
        public LEDApplication Application
        {
            get { return _application; }
            set
            {
                // Make sure we don't have too many children for this new application
                if (value.MaxChildren == 0)
                    _children.Clear();
                else
                {
                    while (_children.Count > value.MaxChildren)
                        _children.RemoveAt(value.MaxChildren);
                }

                // Set the new application
                _application = value;
                _application.PropertyChanged += _application_PropertyChanged;
                OnPropertyChanged("Application");
            }
        }

        [XmlIgnore]
        public LEDApplicationTree Parent { get; private set; }

        private ObservableCollection<LEDApplicationTree> _children = new ObservableCollection<LEDApplicationTree>();
        public ObservableCollection<LEDApplicationTree> Children
        {
            get { return _children; }
            set
            {
                if (_children != null)
                    _children.CollectionChanged -= _children_CollectionChanged;
                _children = value;
                _children.CollectionChanged += _children_CollectionChanged;
                OnPropertyChanged("Children");
            }
        }

        #endregion

        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;
        void OnPropertyChanged(string name)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(name));
        }

        #endregion
    }
}
