﻿using System;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Input;
using System.Xml;
using System.Xml.Serialization;

namespace LEDControl
{
    public class LEDApplicationTreeViewModel : ViewModelBase
    {
        public LEDApplicationTreeViewModel(LEDApplicationTree tree)
        {
            _tree = tree;

            Children = new ObservableCollection<LEDApplicationTreeViewModel>();
            foreach (LEDApplicationTree child in tree.Children)
            {
                Children.Add(new LEDApplicationTreeViewModel(child));
            }
            _tree.PropertyChanged += _tree_PropertyChanged;
            _tree.Children.CollectionChanged += _tree_ChildrenCollectionChanged;
        }

        void _tree_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Application")
                OnPropertyChanged("Application");
        }

        private void _tree_ChildrenCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Reset)
            {
                Children.Clear();
                foreach (LEDApplicationTree child in _tree.Children)
                {
                    Children.Add(new LEDApplicationTreeViewModel(child));
                }
            }

            else
            {
                if (e.OldItems != null)
                {
                    foreach (
                        LEDApplicationTree oldItem in
                            e.OldItems.OfType<LEDApplicationTree>())
                    {
                        LEDApplicationTreeViewModel childVM =
                            Children.FirstOrDefault(
                                item => item._tree == oldItem);
                        if (childVM != null)
                            Children.Remove(childVM);
                    }
                }
                if (e.NewItems != null)
                {
                    foreach (
                        LEDApplicationTree newItem in
                            e.NewItems.OfType<LEDApplicationTree>())
                    {
                        Children.Add(new LEDApplicationTreeViewModel(newItem));
                    }
                }
            }
        }

        public void ChangeApplicationType(string newTypeName)
        {
            LEDApplication newApp = LEDApplication.Create(newTypeName, _tree.Application.Name);

            MessageBoxResult proceed = MessageBoxResult.Yes;
            if (_tree.Children.Count > newApp.MaxChildren)
            {
                if (newApp.MaxChildren == 0)
                    proceed = MessageBox.Show("If you change the type to " + newTypeName +
                        " you will lose the child applications. Do you want to continue?", "Warning",
                        MessageBoxButton.YesNo);
                else
                    proceed = MessageBox.Show("If you change the type to " + newTypeName +
                        " you will lose some of the child applications. Do you want to continue?", "Warning",
                        MessageBoxButton.YesNo);
            }

            if (proceed == MessageBoxResult.Yes)
                _tree.Application = newApp;
        }

        #region Fields

        private readonly LEDApplicationTree _tree;

        #endregion

        #region Commands

        private static void add_Executed(object parameter)
        {
            LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
            LEDApplication menu = LEDApplication.Create(LEDApplication.ApplicationTypes[0], "MENU" + (vm._tree.Parent.Children.Count + 1));
            vm._tree.Parent.AddChild(new LEDApplicationTree(menu));
        }

        private static bool add_CanExecute(object parameter)
        {
            if (parameter is LEDApplicationTreeViewModel)
            {
                LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
                return vm._tree.Parent != null && vm._tree.Parent.Children.Count < vm._tree.Parent.Application.MaxChildren;
            }
            return false;
        }

        private ICommand _addCommand;
        public ICommand AddCommand
        {
            get
            {
                if (_addCommand == null)
                    _addCommand = new RelayCommand(add_Executed, add_CanExecute);
                return _addCommand;
            }
        }

        private static void addChild_Executed(object parameter)
        {
            LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
            LEDApplication menu = LEDApplication.Create(LEDApplication.ApplicationTypes[0], "MENU" + (vm.Children.Count + 1));
            vm._tree.AddChild(new LEDApplicationTree(menu));
        }

        private static bool addChild_CanExecute(object parameter)
        {
            if (parameter is LEDApplicationTreeViewModel)
            {
                LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
                return vm.Children.Count < vm.Application.MaxChildren;
            }
            return false;
        }

        private ICommand _addChildCommand;
        public ICommand AddChildCommand
        {
            get
            {
                if (_addChildCommand == null)
                    _addChildCommand = new RelayCommand(addChild_Executed, addChild_CanExecute);
                return _addChildCommand;
            }
        }

        private static void cut_Executed(object parameter)
        {
            copy_Executed(parameter);
            LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel)parameter;
            vm._tree.Parent.Children.Remove(vm._tree);
        }

        private static bool cut_CanExecute(object parameter)
        {
            return copy_CanExecute(parameter) && delete_CanExecute(parameter);
        }

        private ICommand _cutCommand;
        public ICommand CutCommand
        {
            get
            {
                if (_cutCommand == null)
                    _cutCommand = new RelayCommand(cut_Executed, cut_CanExecute);
                return _cutCommand;
            }
        }

        private static void copy_Executed(object parameter)
        {
            LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel)parameter;
            XmlSerializer xs = new XmlSerializer(typeof(LEDApplicationTree));
            MemoryStream s = new MemoryStream();
            XmlTextWriter xmlTextWriter = new XmlTextWriter(s, Encoding.Unicode);
            xs.Serialize(xmlTextWriter, vm._tree);
            xmlTextWriter.Close();
            s.Close();
            string xml = Encoding.Unicode.GetString(s.GetBuffer());
            for (int i = 0; i < 10; ++i)
            {
                try
                {
                    Clipboard.SetText(xml);
                    break;
                }
                catch (Exception)
                {
                    Thread.Sleep(10);
                }
            }
        }

        private static bool copy_CanExecute(object parameter)
        {
            return (parameter is LEDApplicationTreeViewModel);
        }

        private ICommand _copyCommand;
        public ICommand CopyCommand
        {
            get
            {
                if (_copyCommand == null)
                    _copyCommand = new RelayCommand(copy_Executed, copy_CanExecute);
                return _copyCommand;
            }
        }

        private static void paste_Executed(object parameter)
        {
            LEDApplicationTree tree = GetTreeFromClipboard();
            if (tree != null)
            {
                LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
                if (vm.Application is LEDApplicationSelector)
                {
                    LEDApplicationSelector selector = (LEDApplicationSelector) vm.Application;
                    if (vm.Children.Count + 1 < selector.MaxChildren)
                    {
                        vm._tree.AddChild(tree);
                        return;
                    }
                }

                vm._tree.Parent.AddChild(tree);
            }
        }

        private static bool paste_CanExecute(object parameter)
        {
            return (parameter is LEDApplicationTreeViewModel) &&
                (GetTreeFromClipboard() != null);
        }

        static LEDApplicationTree GetTreeFromClipboard()
        {
            string xml = null;
            for (int i = 0; i < 10; ++i)
            {
                try
                {
                    xml = Clipboard.GetText();
                    break;
                }
                catch (Exception)
                {
                    Thread.Sleep(10);
                }
            }

            if (!string.IsNullOrEmpty(xml) && xml.StartsWith("<"))
            {
                try
                {
                    XmlSerializer xs = new XmlSerializer(typeof(LEDApplicationTree));
                    using (Stream s = new MemoryStream(Encoding.Unicode.GetBytes(xml)))
                    {
                        LEDApplicationTree tree = (LEDApplicationTree) xs.Deserialize(s);
                        return tree;
                    }
                }
                catch { }
            }

            return null;
        }

        private ICommand _pasteCommand;
        public ICommand PasteCommand
        {
            get
            {
                if (_pasteCommand == null)
                    _pasteCommand = new RelayCommand(paste_Executed, paste_CanExecute);
                return _pasteCommand;
            }
        }

        private static void delete_Executed(object parameter)
        {
            LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
            if (MessageBox.Show(string.Format("Are you sure you want to delete {0}?", vm.Application.Name),
                "Confirm Delete", MessageBoxButton.YesNo) == MessageBoxResult.Yes)
            {
                vm._tree.Parent.Children.Remove(vm._tree);
            }
        }

        private static bool delete_CanExecute(object parameter)
        {
            if (parameter is LEDApplicationTreeViewModel)
            {
                LEDApplicationTreeViewModel vm = (LEDApplicationTreeViewModel) parameter;
                return vm._tree.Parent != null && !Runner.Current.IsRunning(vm._tree);
            }
            return false;
        }

        private ICommand _deleteCommand;
        public ICommand DeleteCommand
        {
            get
            {
                if (_deleteCommand == null)
                    _deleteCommand = new RelayCommand(delete_Executed, delete_CanExecute);
                return _deleteCommand;
            }
        }

        #endregion

        #region Properties

        public LEDApplication Application { get { return _tree.Application; } }

        public ObservableCollection<LEDApplicationTreeViewModel> Children { get; private set; }

        #endregion
    }
}
