﻿using System;
using System.Text;

namespace LEDControl
{
    public enum ScrollOption
    {
        None,
        ScrollInFromLeft,
        ScrollInFromRight
    }

    public struct Indicators
    {
        public bool TopLeft;
        public bool BottomLeft;
        public bool TopRight;
        public bool BottomRight;

        public override string ToString()
        {
            return (TopLeft ? "1" : "0") +
                   (BottomLeft ? "1" : "0") +
                   (TopRight ? "1" : "0") +
                   (BottomRight ? "1" : "0");
        }
    }

    [Flags]
    public enum Buttons
    {
        Button1 = 1,
        Button2 = 2,
        Button3 = 4,
        Button4 = 8,
        Button5 = 16,
        Button6 = 32,
        Button7 = 64,
        Button8 = 128
    }

    public class LEDDevice
    {
        #region Public interface

        public LEDDevice()
        {
            deviceChangedWatcher.DeviceChanged += deviceChangedWatcher_DeviceChanged;
            OpenDevice();
        }

        public void Write(string text)
        {
            Write(text, ScrollOption.None);
        }

        public void Write(string text, ScrollOption scrollOption)
        {
            if (currentText != text)
            {
                string command;
                switch (scrollOption)
                {
                    case ScrollOption.None :
                        command = 'M' + text;
                        break;

                    case ScrollOption.ScrollInFromLeft :
                        command = 'R' + text;
                        break;

                    case ScrollOption.ScrollInFromRight :
                        command = 'L' + text;
                        break;

                    default :
                        throw new NotImplementedException(string.Format("Write(..., {0}) is not implemented.",
                                                                        scrollOption));
                }

                SendCommand(command);
                currentText = text;
            }
        }

        public void Refresh()
        {
            Write(currentText);
        }

        public void Erase()
        {
            SendCommand("E");
            currentText = "";
        }

        public void SetIndicators(Indicators indicators)
        {
            SendCommand('I' + indicators.ToString());
            currentIndicators = indicators;
        }

        public void SetBrightness(uint level)
        {
            level = Math.Max(Math.Min(level, 20), 5);
            SendCommand("B" + level);
            currentBrightness = level;
            currentAutoBrightness = false;
        }

        public void SetAutoBrightness()
        {
            SendCommand("Ba");
            currentAutoBrightness = true;
        }

        #region Events

        public event ButtonsChangedHandler ButtonsChanged;
        public event AmbientLightChangedHandler AmbientLightChanged;

        #endregion

        #endregion

        #region Non-public Methods

        private void OpenDevice()
        {
            if (hidDevice == null || !hidDevice.IsOpen)
            {
                hidDevice = HIDDeviceControl.Open(0x4d8, 0xfbd4);
                if (hidDevice != null && hidDevice.IsOpen)
                {
                    hidDevice.DataReceived += hidDevice_DataReceived;
                    if (currentAutoBrightness)
                        SetAutoBrightness();
                    else
                        SetBrightness(currentBrightness);
                    Write(currentText);
                    SetIndicators(currentIndicators);
                }
            }
        }

        private void SendCommand(string command)
        {
            if (hidDevice != null)
            {
                if (command.Length > 62)
                    command = command.Substring(0, 62);
                byte[] bytes = Encoding.ASCII.GetBytes("\x2" + command + '\x3');
                if (bytes.Length <= 64)
                    hidDevice.Write(0, bytes);
            }
        }

        #endregion

        #region Event Handlers

        void deviceChangedWatcher_DeviceChanged(IntPtr wParam, IntPtr lParam)
        {
            if (wParam.ToInt32() == 7)
                OpenDevice();
        }

        void hidDevice_DataReceived(byte[] data)
        {
            if (data.Length == 65 && data[0] == '\0')
            {
                switch ((char) data[1])
                {
                    case 'B' :
                        if (ButtonsChanged != null)
                            ButtonsChanged(this, new ButtonsChangedHandlerArgs {Buttons = (Buttons) data[2]});
                        break;

                    case 'L' :
                        if (AmbientLightChanged != null)
                        {
                            ushort ambientLight = (ushort) (data[2] + (data[3] << 8));
                            AmbientLightChanged(this, new AmbientLightChangedHandlerArgs {AmbientLightLevel = ambientLight});
                        }
                        break;
                }
            }
        }

        #endregion

        #region Fields

        private readonly DeviceChangedWatcher deviceChangedWatcher = new DeviceChangedWatcher();
        private HIDDeviceControl hidDevice;
        private string currentText = "USB ON";
        private Indicators currentIndicators;
        private uint currentBrightness = 20;
        private bool currentAutoBrightness;

        #endregion
    }

    public delegate void ButtonsChangedHandler(object sender, ButtonsChangedHandlerArgs args);

    public class ButtonsChangedHandlerArgs
    {
        public Buttons Buttons;
    }

    public delegate void AmbientLightChangedHandler(object sender, AmbientLightChangedHandlerArgs args);

    public class AmbientLightChangedHandlerArgs
    {
        public ushort AmbientLightLevel;
    }
}
