﻿using System.Collections.Generic;
using System.Linq;

namespace LEDControl
{
    public class LEDDeviceEventArgs
    {
        public LEDApplication OriginalSource;
        public LEDDevice LEDDevice;
        public Runner Runner;
        public bool Handled;
    }

    public class LEDDeviceButtonEventArgs : LEDDeviceEventArgs
    {
        public Buttons Button;
    }

    public class LEDDeviceAmbientLightEventArgs : LEDDeviceEventArgs
    {
        public ushort AmbientLight;
    }

    public class LifecycleArgs : LEDDeviceEventArgs
    {
    }

    public class Runner
    {
        public static void Create(LEDApplicationTree tree, LEDDevice device)
        {
            if (Current == null)
                Current = new Runner(tree, device);
        }

        public static Runner Current { get; set; }

        public bool IsRunning(LEDApplicationTree node)
        {
            return _stack.Contains(node);
        }

        private Runner(LEDApplicationTree tree, LEDDevice device)
        {
            _ledDevice = device;
            LifecycleArgs args = new LifecycleArgs
                                 {
                                     OriginalSource = tree.Application,
                                     LEDDevice = _ledDevice,
                                     Runner = this,
                                 };
            if (tree.Application.Start(tree, args))
            {
                _stack.Add(tree);
                args.Handled = false;
                tree.Application.Activate(tree, args);
            }

            _ledDevice.AmbientLightChanged += _ledDevice_AmbientLightChanged;
            _ledDevice.ButtonsChanged += _ledDevice_ButtonsChanged;
        }

        public void RequestExecute(LEDApplicationTree ledApplicationTree)
        {
            LifecycleArgs args = new LifecycleArgs
            {
                OriginalSource = ledApplicationTree.Application,
                LEDDevice = _ledDevice,
                Runner = this,
                Handled = false,
            };
            if (ledApplicationTree.Application.Start(ledApplicationTree, args))
            {
                _stack.Add(ledApplicationTree);
                args.Handled = false;
                ledApplicationTree.Application.Activate(ledApplicationTree, args);
            }
        }

        public void RequestTerminate(LEDApplication ledApplication)
        {
            if (_stack.Count > 1 && _stack[_stack.Count - 1].Application == ledApplication)
            {
                LifecycleArgs args = new LifecycleArgs
                {
                    OriginalSource = _stack[_stack.Count - 1].Application,
                    LEDDevice = _ledDevice,
                    Runner = this,
                    Handled = false,
                };
                ledApplication.Deactivate(_stack[_stack.Count - 1], args);
                args.Handled = false;
                ledApplication.Terminate(_stack[_stack.Count - 1], args);
                _stack.RemoveAt(_stack.Count - 1);

                args.OriginalSource = _stack[_stack.Count - 1].Application;
                args.Handled = false;
                _stack[_stack.Count - 1].Application.Activate(_stack[_stack.Count - 1], args);
            }
        }

        #region Device Events

        void _ledDevice_ButtonsChanged(object sender, ButtonsChangedHandlerArgs e)
        {
            for (Buttons mask = Buttons.Button1; mask <= Buttons.Button8; mask = (Buttons) ((int) mask << 1))
            {
                if ((mask & _lastButtons) != (mask & e.Buttons))
                {
                    LEDDeviceButtonEventArgs args = new LEDDeviceButtonEventArgs
                                                    {
                                                        OriginalSource = _stack[_stack.Count - 1].Application,
                                                        LEDDevice = _ledDevice,
                                                        Runner = this,
                                                        Handled = false,
                                                        Button = mask
                                                    };
                    if ((mask & e.Buttons) != 0)
                    {
                        for (int index = 0; index < _stack.Count && !args.Handled; ++index)
                        {
                            LEDApplicationTree ledApplicationTree = _stack[index];
                            ledApplicationTree.Application.OnPreviewButtonDown(ledApplicationTree, args);
                        }

                        for (int index = _stack.Count - 1; index >= 0 && !args.Handled; --index)
                        {
                            LEDApplicationTree ledApplicationTree = _stack[index];
                            ledApplicationTree.Application.OnButtonDown(ledApplicationTree, args);
                        }
                    }

                    else
                    {
                        for (int index = 0; index < _stack.Count && !args.Handled; ++index)
                        {
                            LEDApplicationTree ledApplicationTree = _stack[index];
                            ledApplicationTree.Application.OnPreviewButtonUp(ledApplicationTree, args);
                        }

                        for (int index = _stack.Count - 1; index >= 0 && !args.Handled; --index)
                        {
                            LEDApplicationTree ledApplicationTree = _stack[index];
                            ledApplicationTree.Application.OnButtonUp(ledApplicationTree, args);
                        }
                    }
                }
            }

            _lastButtons = e.Buttons;
        }

        void _ledDevice_AmbientLightChanged(object sender, AmbientLightChangedHandlerArgs e)
        {
            if (_stack.Count > 0)
            {
                LEDDeviceAmbientLightEventArgs args = new LEDDeviceAmbientLightEventArgs
                                                      {
                                                          OriginalSource = _stack.Last().Application,
                                                          LEDDevice = _ledDevice,
                                                          Runner = this,
                                                          Handled = false,
                                                          AmbientLight = e.AmbientLightLevel
                                                      };
                for (int index = 0; index < _stack.Count && !args.Handled; ++index)
                {
                    LEDApplicationTree ledApplicationTree = _stack[index];
                    ledApplicationTree.Application.OnPreviewAmbientLightChanged(ledApplicationTree, args);
                }

                for (int index = _stack.Count - 1; index >= 0 && !args.Handled; --index)
                {
                    LEDApplicationTree ledApplicationTree = _stack[index];
                    ledApplicationTree.Application.OnAmbientLightChanged(ledApplicationTree, args);
                }
            }
        }

        #endregion

        #region Properties

        internal LEDApplication CurrentlyRunning
        {
            get
            {
                return _stack.Count > 0 ? _stack.Last().Application : null;
            }
        }

        #endregion

        #region Fields

        private readonly List<LEDApplicationTree> _stack = new List<LEDApplicationTree>();
        private readonly LEDDevice _ledDevice;
        private Buttons _lastButtons;

        #endregion
    }
}
