﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Windows.Data;
using System.Windows;

namespace LEDControl
{
    /// <summary>
    /// Build a list containing the passed-in values. If a value is an IEnumerable, add each
    /// of its items to the list
    /// </summary>
    public class ToListConverter : IMultiValueConverter
    {
        #region IMultiValueConverter Members

        public object Convert(object[] values, Type targetType, object parameter, CultureInfo culture)
        {
            return values.ToList();
        }

        public object[] ConvertBack(object value, Type[] targetTypes, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    [ValueConversion(typeof(object), typeof(Visibility))]
    public class ToVisibilityConverter : IValueConverter
    {
        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            bool visible = false;

            if (value != null)
            {
                if (!(value is bool) || ((bool) value))
                    visible = true;
            }

            if (parameter is string)
            {
                string sparameter = (string) parameter;
                if (sparameter.Contains('i'))
                    visible = !visible;
            }

            return visible ? Visibility.Visible : Visibility.Collapsed;
        }

        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }
    }

    [ValueConversion(typeof(IDictionary), typeof(ICollection))]    
    public class KeyValuePairConverter : IValueConverter
    {
        #region Implementation of IValueConverter

        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            if (value is IDictionary)
                return ((IDictionary) value).Keys;

            return null;
        }

        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    /// <summary>
    /// Given an object (values[0]) and an IList (values[1]) return the index of the 
    /// object in the list. If a third value (values[2]) is provided and is a list,
    /// return the n-th element of that list instead of the index.
    /// </summary>
    public class ToIndexConverter : IMultiValueConverter
    {
        #region Implementation of IMultiValueConverter

        public object Convert(object[] values, Type targetType, object parameter, CultureInfo culture)
        {
            object result = null;
            if (values[1] is IList)
            {
                int index = ((IList) values[1]).IndexOf(values[0]);
                if (values.Length > 2 && values[2] is IList && index != -1 && index < ((IList)values[2]).Count)
                    result = ((IList) values[2])[index];
                else
                    result = index;
            }

            if (parameter is string)
                return string.Format((string) parameter, result);

            return result;
        }

        public object[] ConvertBack(object value, Type[] targetTypes, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    /// <summary>
    /// Given a list of ParameterValues and a parameter name, return the parameter's value
    /// </summary>
    public class ParameterNameToValueConverter : IMultiValueConverter
    {
        #region Implementation of IMultiValueConverter

        public object Convert(object[] values, Type targetType, object parameter, CultureInfo culture)
        {
            if (values[0] is IEnumerable<LEDApplicationDataViewer.DataItem.ParameterValue> &&
                values[1] is DataItemRetriever.Parameter)
            {
                IEnumerable<LEDApplicationDataViewer.DataItem.ParameterValue> parameterValues = (IEnumerable<LEDApplicationDataViewer.DataItem.ParameterValue>) values[0];
                DataItemRetriever.Parameter dirParameter = (DataItemRetriever.Parameter)values[1];
                LEDApplicationDataViewer.DataItem.ParameterValue pv =
                    parameterValues.FirstOrDefault(item => item.ParameterName == dirParameter.Name);
                if (pv != null)
                    return pv.Value;
                
                if (dirParameter.DataType == typeof(string))
                    return "";

                ConstructorInfo ci = dirParameter.DataType.GetConstructor(Type.EmptyTypes);
                if (ci != null)
                    return ci.Invoke(null);
            }

            return null;
        }

        public object[] ConvertBack(object value, Type[] targetTypes, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    public class ToBooleanConverter : IValueConverter
    {
        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            bool result = false;

            if (value != null)
            {
                if (parameter is string && ((string) parameter).Contains('z'))
                {
                    if (value is string && !string.IsNullOrEmpty((string) value))
                        result = true;
                }
            }

            return result;
        }

        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }
    }

    [ValueConversion(typeof(LEDApplicationDataViewer.ButtonFunction.ButtonAction), typeof(string))]
    public class ButtonFunctionActionConverter : IValueConverter
    {
        #region Implementation of IValueConverter

        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            if (value is LEDApplicationDataViewer.ButtonFunction.ButtonAction)
                return Enum.GetName(typeof(LEDApplicationDataViewer.ButtonFunction.ButtonAction), (LEDApplicationDataViewer.ButtonFunction.ButtonAction) value);

            return null;
        }

        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            try
            {
                return Enum.Parse(typeof(LEDApplicationDataViewer.ButtonFunction.ButtonAction), (string) value);
            }
            catch {}

            return LEDApplicationDataViewer.ButtonFunction.ButtonAction.None;
        }

        #endregion
    }
}
